{
****************************************************************
*  Name    : i2cEEPROM.BAS                                     *
*  Author  : Jerry Messina                                     *
*  Date    : 10/18/2018                                        *
*  Version : 1.0                                               *
*  Notes   : I2C EEPROM routines                               *
*          :                                                   *
*  ver 1.0 : initial release                                   *
****************************************************************
}
module i2cEE

// hardware (I2C, I2C2) or software (SI2C) interface selection
#option I2C_EE_INTERFACE = I2C
#if (I2C_EE_INTERFACE = I2C)
  include "I2C.bas"
#elseif (I2C_EE_INTERFACE = I2C2)
  include "I2C2.bas"
#elseif (I2C_EE_INTERFACE = SI2C)
  include "SI2C.bas"
#else
  #error I2C_EE_INTERFACE, "Invalid interface option... must be I2C, I2C2, or SI2C"
#endif

// default eeprom device type
#option I2C_EE_TYPE = 24xx08
// supported eeprom device types
#if isoption(I2C_EE_TYPE) and not(I2C_EE_TYPE in (24xx00, 24xx01, 24xx02, 24xx04, _
     24xx08, 24xx16, 24xx32, 24xx64, 24xx128, 24xx256, 24xx512, 24xx1026, _
     AT24CM01, AT24CM02))
  #error I2C_EE_TYPE, "unsupported I2C_EE_TYPE option: "+I2C_EE_TYPE
#endif
// to add a new device:
//  - add it to the supported 'I2C_EE_TYPE' list above
//  - add/modify the device parameter entries (page size, addr type, size, etc)
//  - modify the Control() and SendAddr() subroutines

//
// eeprom device parameters
// I2C_EE_PAGE/EE_PAGE
//  write page size, set via #option I2C_EE_PAGE since device variants may use 
//  different page sizes (setting #option I2C_EE_PAGE = 0 disables page write code)
// I2C_EEADDR_T
//  typedef used to define the read/write Addr variable, must be large enough to hold
//  the eeprom addr (EE_SIZE-1)
// EE_SIZE
//  eeprom total size, in bytes
// EE_DEVSEL
//  max number of devices on a bus (setting = 0 means a single device)
//  mfgs may have differing support for the number of devices on a bus
//  (hdw addr/sel pins), so be sure to check the datasheet (esp true for ST Micro)
//  changes made to EE_DEVSEL may require changing the SetDevice() and Control() routines
//
// unless noted, the parameters listed are typical for microchip parts
//
#if (I2C_EE_TYPE = 24xx00)      // 128-bit (16x8)
public type I2C_EEADDR_T = byte // addr size
#option I2C_EE_PAGE = 0         // page write size
const
    EE_SIZE = 16,               // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 0               // no device select pins

#elseif (I2C_EE_TYPE = 24xx01)  // 1K-bit (128x8)
public type I2C_EEADDR_T = byte // addr size
#option I2C_EE_PAGE = 8         // page write size
const
    EE_SIZE = 128,              // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 0               // no device select pins

#elseif (I2C_EE_TYPE = 24xx02)  // 2K-bit (256x8)
public type I2C_EEADDR_T = byte // addr size
#option I2C_EE_PAGE = 8         // page write size
const
    EE_SIZE = 256,              // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 0               // no device select pins

#elseif (I2C_EE_TYPE = 24xx04)  // 4K-bit (512x8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 16        // page write size
const
    EE_SIZE = 512,              // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 0               // no device select pins

#elseif (I2C_EE_TYPE = 24xx08)  // 8K-bit (1024x8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 16        // page write size
const
    EE_SIZE = 1024,             // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 2               // 2 devices on a bus (some ST Micro packages)

#elseif (I2C_EE_TYPE = 24xx16)  // 16K-bit (2048x8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 16        // page write size
const
    EE_SIZE = 2048,             // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 0               // no device select pins

#elseif (I2C_EE_TYPE = 24xx32)  // 32K-bit (4096x8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 32        // page write size
const
    EE_SIZE = 4096,             // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 8               // 8 devices on a bus

#elseif (I2C_EE_TYPE = 24xx64) // 64K-bit (8Kx8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 32        // page write size
const
    EE_SIZE = 8192,             // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 8               // 8 devices on a bus

#elseif (I2C_EE_TYPE = 24xx128) // 128K-bit (16Kx8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 64        // page write size
const
    EE_SIZE = 16384,            // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 8               // 8 devices on a bus

#elseif (I2C_EE_TYPE = 24xx256) // 256K-bit (32Kx8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 64        // page write size
const
    EE_SIZE = 32768,            // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 8               // 8 devices on a bus

#elseif (I2C_EE_TYPE = 24xx512) // 512K-bit (64Kx8)
public type I2C_EEADDR_T = word // addr size
#option I2C_EE_PAGE = 128       // page write size
const
    EE_SIZE = 65536,            // eeprom size, in bytes
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 8               // 8 devices on a bus

#elseif (I2C_EE_TYPE = 24xx1026) // 1M-bit (128Kx8)
public type I2C_EEADDR_T = longword // addr size
#option I2C_EE_PAGE = 128       // page write size
const
    EE_SIZE = 131072,           // eeprom size, in bytes (2 blocks of 64K)
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 4               // 4 devices on a bus

#elseif (I2C_EE_TYPE = AT24CM01) // 1M-bit (128Kx8)
public type I2C_EEADDR_T = longword // addr size
#option I2C_EE_PAGE = 256       // page write size
const
    EE_SIZE = 131072,           // eeprom size, in bytes (linear)
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 4               // 4 devices on a bus

#elseif (I2C_EE_TYPE = AT24CM02) // 2M-bit (256Kx8)
public type I2C_EEADDR_T = longword // addr size
#option I2C_EE_PAGE = 256       // page write size
const
    EE_SIZE = 262144,           // eeprom size, in bytes (linear)
    EE_PAGE = I2C_EE_PAGE,
    EE_DEVSEL = 2               // 2 devices on a bus

// ** add new device types here **
#endif      // I2C_EE_TYPE

// WriteProtect WP IO pin control (optional)
// usage example: #option I2C_EE_WP = PORTA.4
#if isoption(I2C_EE_WP)
  #if isvalidportpin(I2C_EE_WP)
    dim EE_WP_PIO as I2C_EE_WP.I2C_EE_WP@
  #else
    #error I2C_EE_WP, "invalid port pin option: "+I2C_EE_WP
  #endif
#endif  // isoption(I2C_EE_WP)

// ack polling timeout for write to complete, in msecs
#option I2C_EE_ACK_TIMEOUT = 10
const EE_ACK_TIMEOUT = I2C_EE_ACK_TIMEOUT

// local working variables...

// publics
public const
    I2C_EEPROM_SIZE = EE_SIZE,
    I2C_EEPROM_PAGESIZE = EE_PAGE
    
public dim 
    Addr as I2C_EEADDR_T,   // eeprom read/write address 
    DevSel as byte          // device select, 0 to (EE_DEVSEL-1)

// I2C addr/control byte lsb (RD_WRN) bit
const R_W = 0

// aliases to interface routines
#if (I2C_EE_INTERFACE = I2C)    // hardware I2C module
  dim I2C_Initialize    as I2C.Initialize
  dim I2C_Start         as I2C.Start
  dim I2C_Stop          as I2C.Stop
  dim I2C_Restart       as I2C.Restart                         
  dim I2C_WriteByte     as I2C.WriteByte
  dim I2C_ReadByte      as I2C.ReadByte
  dim I2C_Ack           as I2C.Acknowledge
  dim NotAcknowledged   as I2C.NotAcknowledged

  // baud rate consts from i2c.bas
  public const
    I2C_100_KHZ = I2C.I2C_100_KHZ,
    I2C_400_KHZ = I2C.I2C_400_KHZ, 
    I2C_1000_KHZ = I2C.I2C_1000_KHZ

#elseif (I2C_EE_INTERFACE = I2C2)    // hardware I2C2 module
  dim I2C_Initialize    as I2C2.Initialize
  dim I2C_Start         as I2C2.Start
  dim I2C_Stop          as I2C2.Stop
  dim I2C_Restart       as I2C2.Restart                         
  dim I2C_WriteByte     as I2C2.WriteByte
  dim I2C_ReadByte      as I2C2.ReadByte
  dim I2C_Ack           as I2C2.Acknowledge
  dim NotAcknowledged   as I2C2.NotAcknowledged

  // baud rate consts from I2C2.bas
  public const
    I2C_100_KHZ = I2C2.I2C_100_KHZ,
    I2C_400_KHZ = I2C2.I2C_400_KHZ, 
    I2C_1000_KHZ = I2C2.I2C_1000_KHZ

#elseif (I2C_EE_INTERFACE = SI2C)   // software SI2C module
  dim I2C_Initialize    as SI2C.Initialize
  dim I2C_Start         as SI2C.Start
  dim I2C_Stop          as SI2C.Stop
  dim I2C_Restart       as SI2C.Restart                         
  dim I2C_WriteByte     as SI2C.WriteByte
  dim I2C_ReadByte      as SI2C.ReadByte
  dim I2C_Ack           as SI2C.Acknowledge
  dim NotAcknowledged   as SI2C.NotAcknowledged
#endif      // I2C_EE_INTERFACE


// clear wdog
private inline sub clrWDT()
    asm
        CLRWDT
    end asm        
end sub

{
****************************************************************************
* Name    : SetAddress                                                     *
* Purpose : Set EEPROM read and write address                              *
****************************************************************************
}
public sub SetAddress(pAddr as I2C_EEADDR_T)
    Addr = pAddr
end sub

public inline function GetAddress() as Addr
end function

{
****************************************************************************
* Name    : SetDevice                                                      *
* Purpose : Set EEPROM device select (if supported), 0 to (EE_DEVSEL-1)    *
****************************************************************************
}                        
public sub SetDevice(pDeviceSel as byte)
    if (EE_DEVSEL > 0) then                // optimize out if EE_DEVSEL=0 (not used)
        if (pDeviceSel < EE_DEVSEL) then
            DevSel = pDeviceSel
        endif
        // if you have physical hardware lines used to select 
        // devices you can add code here to control them
    endif
end sub

{
****************************************************************************
* Name    : Control (private)                                              *
* Purpose : create EEPROM Control byte                                     *
****************************************************************************
}
// eeprom control byte consists of '1010' along with any upper addr and/or
// device select bits, and the i2c R_W bit (which is set to 0)
function Control() as byte
    const I2C_EEPROM = $A0          // eeprom identifier
    dim b as byte                   // control byte
        
    b = I2C_EEPROM                  // 1 0 1 0 x x x R_W
  #if (I2C_EE_TYPE = 24xx00)        // 128-bit (16x8)
    // nothing to add
  #elseif (I2C_EE_TYPE = 24xx01)    // 1K-bit (128x8)
    // nothing to add
  #elseif (I2C_EE_TYPE = 24xx02)    // 2K-bit (256x8)
    // nothing to add
  #elseif (I2C_EE_TYPE = 24xx04)    // 4K-bit (512x8)
    if (Addr.bits(8) = 1) then      // 1 0 1 0 x x B0 R_W
        b.bits(1) = 1                   // B0
    endif        
  #elseif (I2C_EE_TYPE = 24xx08)    // 8K-bit (1024x8)
    if (Addr.bits(8) = 1) then      // 1 0 1 0 x B1 B0 R_W
        b.bits(1) = 1                   // B0
    endif        
    if (Addr.bits(9) = 1) then
        b.bits(2) = 1                   // B1
    endif
    if (DevSel = 1) then
        b.bits(3) = 1                   // hardware addr A2 (ST M24C08)
    endif        
  #elseif (I2C_EE_TYPE = 24xx16)    // 16K-bit (2048x8)
    if (Addr.bits(8) = 1) then      // 1 0 1 0 B2 B1 B0 R_W
        b.bits(1) = 1                   // B0
    endif        
    if (Addr.bits(9) = 1) then
        b.bits(2) = 1                   // B1
    endif        
    if (Addr.bits(10) = 1) then
        b.bits(3) = 1                   // B2
    endif        
  #elseif (I2C_EE_TYPE = 24xx32)    // 32K-bit (4096x8)
    b = b or (DevSel << 1)          // 1 0 1 0 A2 A1 A0 R_W
  #elseif (I2C_EE_TYPE = 24xx64)    // 64K-bit (8Kx8)
    b = b or (DevSel << 1)          // 1 0 1 0 A2 A1 A0 R_W
  #elseif (I2C_EE_TYPE = 24xx128)   // 128K-bit (16Kx8)
    b = b or (DevSel << 1)          // 1 0 1 0 A2 A1 A0 R_W
  #elseif (I2C_EE_TYPE = 24xx256)   // 256K-bit (32Kx8)
    b = b or (DevSel << 1)          // 1 0 1 0 A2 A1 A0 R_W
  #elseif (I2C_EE_TYPE = 24xx512)   // 512K-bit (64Kx8)
    b = b or (DevSel << 1)          // 1 0 1 0 A2 A1 A0 R_W
  #elseif (I2C_EE_TYPE = 24xx1026)  // 1M-bit (128Kx8)
    if (Addr.bits(16) = 1) then     // 1 0 1 0 A2 A1 B0 R_W
        b.bits(1) = 1                   // B0 block select (upper/lower 64K)
    endif        
    b = b or (DevSel << 2)              // A1, A2 chip select bits
  #elseif (I2C_EE_TYPE = AT24CM01)  // 1M-bit (128Kx8)
    if (Addr.bits(16) = 1) then     // 1 0 1 0 A2 A1 P0 R_W
        b.bits(1) = 1                   // P0 block select (upper/lower 64K)
    endif        
    b = b or (DevSel << 2)              // A1, A2 chip select bits
  #elseif (I2C_EE_TYPE = AT24CM02)  // 2M-bit (256Kx8)
    if (Addr.bits(16) = 1) then     // 1 0 1 0 A2 A17 A16 R_W
        b.bits(1) = 1                   // A16
    endif        
    if (Addr.bits(17) = 1) then
        b.bits(2) = 1                   // A17
    endif
    if (DevSel = 1) then
        b.bits(3) = 1                   // hardware addr A2
    endif        
  // ** add new device types here **
  #endif    // I2C_EE_TYPE
    result = b
end function

{
****************************************************************************
* Name    : SendAddr (private)                                             *
* Purpose : Send EEPROM address                                            *
*         : note- this routine does not send any of the upper addr/dev sel *
*         : bits that are part of the control byte                         *
****************************************************************************
}
sub SendAddr()
    // send addr byte(s)
  #if (I2C_EE_TYPE = 24xx00)        // 128-bit (16x8)
    I2C_WriteByte(Addr)    
  #elseif (I2C_EE_TYPE = 24xx01)    // 1K-bit (128x8)
    I2C_WriteByte(Addr)    
  #elseif (I2C_EE_TYPE = 24xx02)    // 2K-bit (256x8)
    I2C_WriteByte(Addr)    
  #elseif (I2C_EE_TYPE = 24xx04)    // 4K-bit (512x8)
    I2C_WriteByte(Addr.byte0)    
  #elseif (I2C_EE_TYPE = 24xx08)    // 8K-bit (1024x8)
    I2C_WriteByte(Addr.byte0)    
  #elseif (I2C_EE_TYPE = 24xx16)    // 16K-bit (2048x8)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = 24xx32)    // 32K-bit (4096x8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = 24xx64)    // 64K-bit (8Kx8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = 24xx128)   // 128K-bit (16Kx8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = 24xx256)   // 256K-bit (32Kx8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = 24xx512)   // 512K-bit (64Kx8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = 24xx1026)  // 1M-bit (128Kx8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = AT24CM01)  // 1M-bit (128Kx8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  #elseif (I2C_EE_TYPE = AT24CM02)  // 2M-bit (256Kx8)
    I2C_WriteByte(Addr.byte1)
    I2C_WriteByte(Addr.byte0)
  // ** add new device types here **
  #endif    // I2C_EE_TYPE
end sub

{
****************************************************************************
* Name    : WriteProtect                                                   *
* Purpose : WP pin control (optional)                                      *
*         : WP=0  normal read/write operation                              *
*         : WP=1  write operations are inhibited                           *
****************************************************************************
}
#if isoption(I2C_EE_WP)
public const
    WP_ON = 1,
    WP_OFF = 0
    
public sub WriteProtect(onoff as byte)
    if (onoff = WP_OFF) then    // write enabled
        low(EE_WP_PIO)
    else                        // write disabled
        high(EE_WP_PIO)
    endif
end sub

public function WriteProtect() as byte
    if (EE_WP_PIO = 0) then
        result = 0
    else
        result = 1
    endif
end function    
#endif  // isoption(I2C_EE_WP)

{
****************************************************************************
* Name    : IsPresent                                                      *
* Purpose : Send control byte and see if device ACKs                       *
*         : returns true (device ACKed) or false (device NACKed)           *
*         : can be used for device presence check (leave timeout=0)        *
*         : or a write busy check with a 5-10ms typ timeout                *
****************************************************************************
}
public function IsPresent(pTimeout as byte=0) as boolean
    const POLL_INTERVAL = 100   // interval for polling, in usecs (1000 max)
    dim tout as word            // timeout increments
    
    // convert msec pTimeout parameter to POLL_INTERVAL steps
    // (typ faster than just using 1ms intervals)
    tout = pTimeout * (1000/POLL_INTERVAL)
    // send control byte until device returns ACK (or timeout)
    repeat
        I2C_Start()
        I2C_WriteByte(Control())
        I2C_Stop()
        if (bit(NotAcknowledged) = 0) then   // 0=ACK was received
            result = true
            exit
        endif
        if (tout > 0) then
            delayus(POLL_INTERVAL)
            tout = tout - 1
        endif
    until (tout = 0)
    // device did not ack its addr within the timeout
    result = false
end function

{
****************************************************************************
* Name    : ReadByte                                                       *
* Purpose : Read a byte from I2C EEPROM                                    *
****************************************************************************
} 
public function ReadByte() as byte
    dim ctrl as byte
    
    ctrl = Control()                // control byte with R_W=0 (write)
    I2C_Start()
    I2C_WriteByte(ctrl)             // random read... send ctrl
    SendAddr()                      // send addr byte(s)
    I2C_Restart()
    ctrl.bits(R_W) = 1              // set read mode                
    I2C_WriteByte(ctrl)             // control byte with R_W=1 (read)
    result = I2C_ReadByte()
    I2C_Ack(I2C_NOT_ACKNOWLEDGE)
    I2C_Stop()
    Addr = Addr + 1                 // increment address
end function

{
****************************************************************************
* Name    : WriteByte                                                      *
* Purpose : Write a byte to I2C EEPROM                                     *
****************************************************************************
} 
public sub WriteByte(b as byte)
    I2C_Start()
    I2C_WriteByte(Control())    // control byte with R_W=0 (write)
    SendAddr()                  // send addr byte(s)
    I2C_WriteByte(b)
    I2C_Stop()
    IsPresent(EE_ACK_TIMEOUT)   // wait for write to complete
    Addr = Addr + 1             // increment address
end sub

{
****************************************************************************
* Name    : ReadBoolean                                                    *
* Purpose : Read a boolean from the I2C EEPROM                             *
****************************************************************************
}
public function ReadBoolean() as boolean
    result = boolean(ReadByte)
end function 
{
****************************************************************************
* Name    : ReadWord                                                       *
* Purpose : Read a word from the I2C EEPROM                                *
****************************************************************************
}
public function ReadWord() as word
    result.Bytes(0) = ReadByte
    result.Bytes(1) = ReadByte
end function  
{
****************************************************************************
* Name    : ReadLongWord                                                   *
* Purpose : Read a long word from the I2C EEPROM                           *
****************************************************************************
}
public function ReadLongWord() as longword
    result.Bytes(0) = ReadByte
    result.Bytes(1) = ReadByte
    result.Bytes(2) = ReadByte
    result.Bytes(3) = ReadByte
end function  
{
****************************************************************************
* Name    : ReadFloat                                                      *
* Purpose : Read a float from the I2C EEPROM                               *
****************************************************************************
}
public function ReadFloat() as float
    result.Bytes(0) = ReadByte
    result.Bytes(1) = ReadByte
    result.Bytes(2) = ReadByte
    result.Bytes(3) = ReadByte
end function 
{
****************************************************************************
* Name    : ReadStr                                                        *
* Purpose : Read a string from the I2C EEPROM                              *
****************************************************************************
}
public sub ReadStr(byref pText as string)
    dim i, b as byte
    dim ctrl as byte
    
    ctrl = Control()
    I2C_Start()
    I2C_WriteByte(ctrl)             // send control byte with R_W=0
    SendAddr()                      // send addr byte(s)
    I2C_Restart()
    ctrl.bits(R_W) = 1              // set read mode                
    I2C_WriteByte(ctrl)             // control byte with R_W=1 (read)
    b = I2C_ReadByte()
    i = 0
    while (b <> 0)
        pText(i) = b
        I2C_Ack(I2C_ACKNOWLEDGE)    // ack previous I2C_ReadByte
        b = I2C_ReadByte()
        i = i + 1
    end while
    I2C_Ack(I2C_NOT_ACKNOWLEDGE)    // nack last byte
    I2C_Stop()
    pText(i) = null
    inc(Addr, i + 1)
end sub
{
****************************************************************************
* Name    : ReadArray                                                      *
* Purpose : Read a byte array from the I2C EEPROM                          *
****************************************************************************
}
public sub ReadArray(byref pArray() as byte, pMax as word = $FFFF)
    dim i as word
    dim ctrl as byte

    ctrl = Control()
    I2C_Start()
    I2C_WriteByte(ctrl)             // send control byte with R_W=0
    SendAddr()                      // send addr byte(s)
    I2C_Restart()
    ctrl.bits(R_W) = 1              // set read mode                
    I2C_WriteByte(ctrl)             // control byte with R_W=1 (read)
   
    i = 1
    pArray(0) = I2C_ReadByte()
    while (i < pMax) and (i <= bound(pArray))
        I2C_Ack(I2C_ACKNOWLEDGE)    // ack previous I2C_ReadByte
        pArray(i) = I2C_ReadByte()
        i = i + 1
    end while
    I2C_Ack(I2C_NOT_ACKNOWLEDGE)    // nack last byte
    I2C_Stop()
    Addr = Addr + i
end sub 
{
****************************************************************************
* Name    : WriteBoolean                                                   *
* Purpose : Write a boolean value to the I2C EEPROM                        *
****************************************************************************
}
public sub WriteBoolean(pValue as boolean)
    WriteByte(byte(pValue))
end sub
{
****************************************************************************
* Name    : WriteWord                                                      *
* Purpose : Write a word value to the I2C EEPROM                           *
****************************************************************************
}
public sub WriteWord(pValue as word)
    WriteByte(pValue.Bytes(0))
    WriteByte(pValue.Bytes(1))
end sub
{
****************************************************************************
* Name    : WriteLongWord                                                  *
* Purpose : Write a word value to the I2C EEPROM                           *
****************************************************************************
}
public sub WriteLongWord(pValue as longword) 
    WriteByte(pValue.Bytes(0))
    WriteByte(pValue.Bytes(1))
    WriteByte(pValue.Bytes(2))
    WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : WriteFloat                                                     *
* Purpose : Write a floating point number to the I2C EEPROM                *
****************************************************************************
}
public sub WriteFloat(pValue as float) 
    WriteByte(pValue.Bytes(0))
    WriteByte(pValue.Bytes(1))
    WriteByte(pValue.Bytes(2))
    WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : WriteStr                                                       *
* Purpose : Write a string value to the I2C EEPROM                         *
****************************************************************************
}
public sub WriteStr(byref pText as string)
    dim i, b as byte  

    clrWDT()
    i = 0
    repeat
        b = pText(i)
        WriteByte(b)
        i = i + 1
    until (b = 0)   // terminate on end of string null char
end sub

// write a const string ie WriteStr(@("THIS IS A CONST STRING"))
inline function TBLRD_POSTINC() as TABLAT
    asm
        TBLRD*+
    end asm
end function

public sub WriteStr(pText as TABLEPTR)
    clrWDT()
    repeat
        WriteByte(TBLRD_POSTINC())
    until (TABLAT = 0)
end sub

{
****************************************************************************
* Name    : WriteArray                                                     *
* Purpose : Write a byte array to the I2C EEPROM                           *
*         : (see WritePage for a faster version)                           *
****************************************************************************
}
public sub WriteArray(byref pArray() as byte, pMax as word = $FFFF)
    dim i as word
   
    i = 0
    while (i < pMax) and (i <= bound(pArray))
        clrWDT()
        WriteByte(pArray(i))
        i = i + 1
    end while
end sub

{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a boolean from the I2C EEPROM                             *
****************************************************************************
}
sub ReadItem(byref pValue as boolean)
    pValue = boolean(ReadByte)
end sub 
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a byte from the I2C EEPROM                                *
****************************************************************************
}
sub ReadItem(byref pValue as byte)
    pValue = ReadByte
end sub  
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a shortint from the I2C EEPROM                            *
****************************************************************************
}
sub ReadItem(byref pValue as shortint)
    pValue = ReadByte
end sub  
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a word from the I2C EEPROM                                *
****************************************************************************
}
sub ReadItem(byref pValue as word)
    pValue = ReadWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read an integer from the I2C EEPROM                            *
****************************************************************************
}
sub ReadItem(byref pValue as integer)
    pValue = ReadWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long word from the I2C EEPROM                           *
****************************************************************************
}
sub ReadItem(byref pValue as longword)
    pValue = ReadLongWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long integer from the I2C EEPROM                        *
****************************************************************************
}
sub ReadItem(byref pValue as longint)
    pValue = ReadLongWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a floating point number from the I2C EEPROM               *
****************************************************************************
}
sub ReadItem(byref pValue as float) 
    pValue = ReadFloat
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a string from the I2C EEPROM                              *
****************************************************************************
}
sub ReadItem(byref pText as string)
    ReadStr(pText)
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a byte array from the I2C EEPROM                          *
****************************************************************************
}
sub ReadItem(byref pArray() as byte)
    ReadArray(pArray)
end sub
{
****************************************************************************
* Name    : Read (COMPOUND)                                                *
* Purpose : Read an item from the I2C EEPROM                               *
****************************************************************************
}
public compound sub Read(SetAddress, ReadItem)

{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a boolean value to the I2C EEPROM                        *
****************************************************************************
}
sub WriteItem(pValue as boolean)
    WriteByte(byte(pValue))
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a byte value to the I2C EEPROM                           *
****************************************************************************
}
sub WriteItem(pValue as WREG)
    WriteByte(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a short int value to the I2C EEPROM                      *
****************************************************************************
}
sub WriteItem(pValue as shortint)
    WriteByte(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a word value to the I2C EEPROM                           *
****************************************************************************
}
sub WriteItem(pValue as word) 
    WriteWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write an integer value to the I2C EEPROM                       *
****************************************************************************
}
sub WriteItem(pValue as integer) 
    WriteWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long word to the I2C EEPROM                            *
****************************************************************************
}
sub WriteItem(pValue as longword) 
    WriteLongWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long integer to the I2C EEPROM                         *
****************************************************************************
}
sub WriteItem(pValue as longint) 
    WriteLongWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a floating point number to the I2C EEPROM                *
****************************************************************************
}
sub WriteItem(pValue as float) 
    WriteByte(pValue.Bytes(0))
    WriteByte(pValue.Bytes(1))
    WriteByte(pValue.Bytes(2))
    WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a string value to the I2C EEPROM                         *
****************************************************************************
}
sub WriteItem(pText as string)
    WriteStr(pText)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a byte array to the I2C EEPROM                           *
****************************************************************************
}
sub WriteItem(byref pArray() as byte)
    WriteArray(pArray)
end sub
{
****************************************************************************
* Name    : Write (COMPOUND)                                               *
* Purpose : Write an item to the I2C EEPROM                                *
****************************************************************************
}
public compound sub Write(SetAddress, WriteItem)

{
****************************************************************************
* Name    : Fill                                                           *
* Purpose : Fill EEPROM with value                                         *
****************************************************************************
}
// if I2C_EE_PAGE > 0 then page writes are used which saves significant time
// for example, filling a 1024 byte device takes about 300ms using page writes
// vs 3.5secs using individual byte writes
public sub Fill(pVal as byte = $FF)
  #if (I2C_EE_PAGE = 0)     // if page write not supported then just write bytes
    dim count as I2C_EEADDR_T

    SetAddress(0)
    for count = 0 to (EE_SIZE-1)
        clrWDT()
        WriteByte(pVal)
    next
  #else                     // use page write
    dim count, pages as word

    SetAddress(0)
    pages = EE_SIZE/EE_PAGE         // number of pages
    repeat
        clrWDT()
        I2C_Start()                 // fill one page with pVal
        I2C_WriteByte(Control())
        SendAddr()
        count = EE_PAGE
        repeat
            I2C_WriteByte(pVal)
            count = count - 1
        until (count = 0)
        I2C_Stop()
        IsPresent(EE_ACK_TIMEOUT)   // wait for write to complete
        Addr = Addr + EE_PAGE       // next page addr
        pages = pages - 1
    until (pages = 0)    
  #endif    // I2C_EE_PAGE
end sub

{
****************************************************************************
* Name    : WritePage                                                      *
* Purpose : Write an array to EEPROM using Page Write (if supported)       *
*         : significantly faster than WriteArray for large data sets       *
*         : USES FSR2                                                      *
****************************************************************************
}
public sub WritePage(pAddr as I2C_EEADDR_T, byref pArray() as byte, pCount as word)
  #if (I2C_EE_PAGE = 0)     // if page write not supported then just write array
    SetAddress(pAddr)
    WriteArray(pArray, pCount)
  #else                     // use page write
    dim bcnt as word,       // number of bytes to write
        i as word
 
    // make sure request is valid
    if ((longword(pAddr) + pCount) > EE_SIZE) then
        exit
    endif

    // set initial address
    SetAddress(pAddr)
    // set a pointer to the source data
    FSR2 = addressof(pArray())

    // ensure the first write doesn't cross a page boundary (otherwise data will wrap)
    // after that we'll be page-aligned and we can write full pages
    bcnt = pAddr and (EE_PAGE-1)        // mask all but in-page addr bits
    bcnt = EE_PAGE - bcnt               // num bytes to end of page

    // write page(s)
    while (pCount <> 0)
        // limit number of bytes to remaining count
        if (bcnt > pCount) then
            bcnt = pCount
        endif
        // write one page
        clrWDT()
        I2C_Start()
        I2C_WriteByte(Control())
        SendAddr()
        // send page data
        i = bcnt                    // we need bcnt later, so make a copy
        while (i <> 0)
            I2C_WriteByte(POSTINC2)
            i = i - 1
        end while
        I2C_Stop()
        // ack polling for write to complete
        IsPresent(EE_ACK_TIMEOUT)   // wait for write to complete
        // adjust counts
        Addr = Addr + bcnt          // increment eeprom address
        pCount = pCount - bcnt      // decr total count 
        bcnt = EE_PAGE              // assume we'll write a whole page  
    end while
  #endif    // I2C_EE_PAGE
end sub

{
****************************************************************************
* Name    : Initialize                                                     *
* Purpose : Initialize I2C EEPROM                                          *
****************************************************************************
}
#if (I2C_EE_INTERFACE = I2C) or (I2C_EE_INTERFACE = I2C2)
public sub Initialize(clk_rate as byte=I2C_100_KHZ)
    // init I2C bus
    I2C_Initialize(clk_rate)
    // module initialization
    Addr = 0
    DevSel = 0
end sub
#elseif (I2C_EE_INTERFACE = SI2C)
public sub Initialize()
    // init I2C bus
    I2C_Initialize()
    // module initialization
    Addr = 0
    DevSel = 0
end sub
#endif

// set I2C bus to a known state by sending START-STOP sequence
public sub InitBus()
    delayus(5)
    I2C_Start()
    delayus(10)
    I2C_Stop()
    delayus(5)
end sub

end module
